"""
Copyright (c) 2008, Carl Scheffler
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * The name of the copyright owner may not be used to endorse or promote
      products derived from this software without specific prior written
      permission.

THIS SOFTWARE IS PROVIDED BY <copyright holder> ``AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <copyright holder> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

"""
Last modified: 3 April 2006

FabSim: Functions for simulating the FabLab machines. Each of the functions below
        opens a Tk window and animates a toolpath.

animate_paths - For the internal _paths_ data format
animate_epi   - For Epilog laser cutter toolpaths
animate_camm  - For Camm vinyl cutter toolpaths
"""

from __future__ import division
import Tkinter

def animate_paths(paths, scale = 40, window = (1000, 650), title = '',\
                  grid = None, format_callback = None, speed_callback = None):
    """
    paths: internal paths structure
    scale: all values in paths multiplied by this, seeing as internal units are inches,
           this is simply the dpi.
    window: size of the window on screen
    title: window title
    grid: a 2-tuple containing the distances between the x and y gridlines
    format_callback: defines colours for drawing lines.
                     takes 1 argument, the path parameters dict.
                     returns a colour code of the form '#0369CF'
    speed_callback: adjust the speed of the drawing.
                    takes 1 argument, the path parameters dict.
                    returns a float with which to multiply the speed.
    """
    if format_callback is None:
        format_callback = lambda x: '#000000'

    if speed_callback is None:
        speed_callback = lambda x: 1.0

    animate_params = {
        'path_index': 0,    # current path index
        'segment_index': 0, # current segment index inside path
        'distance': 0,      # distance already travelled along current segment
    }
    def animate():
        # Check if path completed on this tick
        #   yes: compute overflow into next path
        #        delete "busy" lines from canvas
        #        plot full path into "done"
        # Compute end-point of current path and plot into "busy"
        import math

        to_travel = 0.4 # inches
        while True:
            if animate_params['path_index'] >= len(paths): # Done
                root.update_idletasks()
                return
            
            path = paths[animate_params['path_index']]
            params = path['params']
            path = path['path']

            format = format_callback(params)

            # If completed this path, rerender and go to the next
            if animate_params['segment_index'] >= len(path)-1:
                canvas.delete("busy")
                coords = []
                for i in range(len(path)):
                    for j in range(2):
                        coords.append(path[i][j]*scale)
                canvas.create_line(coords, tag="done", fill=format)
                
                animate_params['path_index'] += 1
                animate_params['segment_index'] = 0
                continue

            speed_factor = speed_callback(params)
            to_travel *= speed_factor
            s = animate_params['segment_index']
            coords = [path[s][0], path[s][1], path[s+1][0], path[s+1][1]]
            seg_length = math.hypot(coords[0]-coords[2], coords[1]-coords[3])

            # Check if the whole segment should be completed
            if seg_length < animate_params['distance']+to_travel:
                for i in range(4): coords[i] *= scale
                canvas.create_line(coords, tag="busy", fill=format)
                animate_params['segment_index'] += 1
                to_travel -= seg_length - animate_params['distance']
                to_travel /= speed_factor
                animate_params['distance'] = 0

                continue
            
            # Draw only part of the segment
            frac = (animate_params['distance']+to_travel) / seg_length
            for i in range(2):
                coords[i+2] = coords[i] + (coords[i+2]-coords[i])*frac
            for i in range(4): coords[i] *= scale
            canvas.create_line(coords, tag="busy", fill=format)
            animate_params['distance'] += to_travel
            
            # Render and schedule next update
            root.update_idletasks()
            root.after(50, animate)
            return
        
    # Initialize canvas
    root = Tkinter.Tk()
    root.title(title)

    canvasframe = Tkinter.Frame(root)
    xscrollbar = Tkinter.Scrollbar(canvasframe, orient=Tkinter.HORIZONTAL)
    xscrollbar.grid(row=1, column=0, sticky=Tkinter.E+Tkinter.W)
    yscrollbar = Tkinter.Scrollbar(canvasframe)
    yscrollbar.grid(row=0, column=1, sticky=Tkinter.N+Tkinter.S)

    canvas = Tkinter.Canvas(canvasframe, width=window[0], height=window[1], background='white',
                            xscrollcommand=xscrollbar.set, yscrollcommand=yscrollbar.set)
    canvas.grid(row=0, column=0, sticky=Tkinter.N+Tkinter.S+Tkinter.E+Tkinter.W)
    canvas.configure(scrollregion=(0,0,window[0],window[1]))
    xscrollbar.config(command=canvas.xview)
    yscrollbar.config(command=canvas.yview)

    canvasframe.pack()
    
    # Find bounding box
    max_xy = [-1, -1]
    for path in paths:
        for point in path['path']:
            for i in range(2):
                xy = point[i] * scale
                if xy > max_xy[i]: max_xy[i] = xy
    max_xy[0] = max(max_xy[0], window[0])
    max_xy[1] = max(max_xy[1], window[1])
    
    # Draw grid lines
    if grid != None:
        for x in range(0, max_xy[0], grid[0]*scale):
            canvas.create_line([x, 0, x, max_xy[1]], tag="grid", dash=4, fill="#F0E030")
        for y in range(0, max_xy[1], grid[1]*scale):
            canvas.create_line([0, y, max_xy[0], y], tag="grid", dash=4, fill="#F0E030")

    # Setup scroll bars
    xscrollmin = 0
    xscrollmax = max_xy[0]
    yscrollmin = 0
    yscrollmax = max_xy[1]
    canvas.configure(scrollregion = (xscrollmin, yscrollmin, xscrollmax, yscrollmax))
    if xscrollmax == window[0]:
        xscrollbar.grid_forget()
    else:
        xscrollbar.grid(row = 1, column = 0, sticky = Tkinter.E+Tkinter.W)
    if yscrollmax == window[1]:
        yscrollbar.grid_forget()
    else:
        yscrollbar.grid(row = 0, column = 1, sticky = Tkinter.N+Tkinter.S)

    animate()
    root.update()
    
    return root

def animate_epi(commands, dpi = 40, window = (960, 480), title = 'Epilog simulation',\
                grid = (1, 1), format_callback = None, speed_callback = None):

    from convert import epi2paths
    
    def default_callback(params):
        maxpower = 100
        
        if params['power'] > maxpower:
            return "#000000"
        else:
            return "#%02x%02x%02x"%(((maxpower-params['power'])*192//maxpower,)*3)
            
    if format_callback is None:
        format_callback = default_callback

    if speed_callback is None:
        speed_callback = lambda params: params['speed']/100.

    return animate_paths(epi2paths(commands), scale = dpi, window = window,\
                         title = title, grid = grid, format_callback = format_callback,\
                         speed_callback = speed_callback)

def animate_camm(commands, dpi = 40, window = (960, 480), title = 'Camm simulation',\
                 grid = (1, 1), format_callback = None, speed_callback = None):

    from convert import camm2paths
    
    def default_callback(params):
        maxforce = 200

        if params['force'] > maxforce:
            colour = "#000000"
        else:
            colour = "#%02x%02x%02x"%(((maxforce-params['force'])*192//maxforce,)*3)
        return colour
            
    if format_callback is None:
        format_callback = default_callback

    if speed_callback is None:
        speed_callback = lambda params: params['speed']/100.

    return animate_paths(camm2paths(commands), scale = dpi, window = window,\
                         title = title, grid = grid, format_callback = format_callback,\
                         speed_callback = speed_callback)
